
#include "problem.h"
#include "preprocess.h"
#include "parser.h"

int main(int argc, char *argv[]) {
  HSPS::Statistics stats;
  stats.enable_interrupt();
  stats.start();

  HSPS::StringTable symbols(50, HSPS::uppercase_map);
  HSPS::Parser* reader = new HSPS::Parser(symbols);

  HSPS::PDDL_Base::warning_level = 0;
  for (int k = 1; k < argc; k++) {
    if ((strcmp(argv[k],"-v") == 0) && (k < argc - 1)) {
      HSPS::Instance::default_trace_level = atoi(argv[++k]);
    }
    else if (*argv[k] != '-') {
      reader->read(argv[k], false);
    }
  }

  reader->post_process();

  HSPS::Instance instance;
  HSPS::Preprocessor prep(instance, stats);
  HSPS::Preprocessor::remove_static_atoms = false;

  stats.start();
  reader->instantiate(instance);
  prep.preprocess(false);
  stats.stop();
  std::cerr << "instantiation finished in " << stats.time() << " seconds"
	    << std::endl;

  for (HSPS::index_type k = 0; k < instance.n_atoms(); k++) {
    HSPS::PDDL_Name* name =
      (HSPS::PDDL_Name*)instance.atoms[k].name->cast_to("PDDL_Name");
    if (name != 0) {
      if ((strncmp(name->symbol()->print_name, "JUSTIFIED-", 10) == 0) &&
	  !instance.atoms[k].del_by.empty()) {
	instance.atoms[k].init = true;
	instance.atoms[k].goal = true;
      }
      else if (name->is_neg() && instance.atoms[k].init) {
	instance.atoms[k].init = false;
      }
    }
  }

  HSPS::Instance::write_metric = false;
  reader->write_problem_begin(std::cout);
  reader->write_objects(std::cout, true);
  std::cout << std::endl << "  (:init";
  for (HSPS::index_type k = 0; k < instance.n_atoms(); k++)
    if (instance.atoms[k].init) {
      std::cout << std::endl << "    " << instance.atoms[k].name;
    }
  for (HSPS::index_type k = 0; k < reader->dom_init.length(); k++)
    if (!reader->dom_init[k]->pred->modded) {
      std::cout << " (" << reader->dom_init[k]->pred->print_name;
      ((HSPS::PDDL_Base::AtomBase*)reader->dom_init[k])->print(std::cout);
      std::cout << ")";
    }
  std::cout << ")" << std::endl << std::endl << "  (:goal (and";
  for (HSPS::index_type k = 0; k < instance.n_atoms(); k++)
    if (instance.atoms[k].goal) {
      std::cout << std::endl << "    " << instance.atoms[k].name;
    }
  std::cout << "))" << std::endl << std::endl;
  std::cout << ")" << std::endl;
}
