
(load "/home/path/pkg/inval/inval.lsp")
(setq *pddl-keywords* (reassoc 'actors ':actors *pddl-keywords*))
(setq *allow-delegation-of-intention* nil)

(setq input-file-name "ry-story.npddl")
(setq output-file-name "ry-story-jtc.pddl")
(setq intendable-predicates
      '(married-to loves controls has dead at))
(setq delegatable-predicates
      '(loves has dead at))
(setq dg '((married-to loves at)
	   (loves controls)
	   (controls has)
	   (has has dead at)
	   (dead at)))

;; (setq input-file-name "ex-story.npddl")
;; (setq output-file-name "ex-story-jtc.pddl")
;; (setq intendable-predicates
;;       '(married-to loves controls has dead at))
;; (setq delegatable-predicates
;;       '(loves has dead at))
;; (setq dg '((married-to loves at)
;; 	   (loves controls)
;; 	   (controls has)
;; 	   (has has dead at loves)
;; 	   (dead at loves)))


(defun basic-achievable-intentions (actions)
  (let ((res nil))
    (dolist (act actions)
      (let ((adds (collect-adds (cdr (assoc ':effect (cdr act)))))
	    (actors (cdr (assoc ':actors (cdr act)))))
	(cond
	 (actors
	  (dolist (add adds)
	    (cond
	     ((not (eq (car (first add)) 'intends))
	      (setq res (add-to-set
			 (cons (car (first add))
			       (remove-duplicates
				(action-parameter-list-to-type-list
				 actors act)))
			 res)))
	     ))))))
    res))

(defun action-parameter-list-to-type-list (params act)
  (let ((decls (cdr (assoc ':parameters (cdr act)))))
    (mapcar #'(lambda (p)
		(let ((ptype (assoc p decls)))
		  (cond (ptype (cdr ptype))
			(t (error "~a is not a parameter of ~a" p act)))))
	    params)))

(defun dg-reachable (dg set new)
  (cond ((endp new) set)
	(t (dg-reachable
	    dg (remove-duplicates (append set (cdr (assoc (first new) dg))))
	    (append (rest new)
		    (remove-if #'(lambda (x) (find x set)) 
			       (cdr (assoc (first new) dg))))))
	))

(defun rename-predicate-parameters (pred &optional (level nil))
  (cons (car pred)
	(mapcar
	 #'(lambda (a)
	     (cond (level
		    (cons (symnumcat (car a) '- (car pred) '- level)
			  (cdr a)))
		   (t (cons (symnumcat (car a) '- (car pred)) (cdr a)))))
	 (cdr pred))))

(defun strip-types (pred)
  (cons (car pred) (mapcar #'car (cdr pred))))

(defun instantiate-fact-parameters (actions predicates)
  (cond
   ((endp actions) nil)
   (t (let ((fps (rassoc-all 'fact (cdr (assoc ':parameters
					       (cdr (first actions)))))))
	(cond
	 ((null fps)
	  (cons (first actions)
		(instantiate-fact-parameters (rest actions) predicates)))
	 ((> (length fps) 1)
	  (error "actions with >1 fact parameters not supported: ~s"
		 (first actions)))
	 (t (let ((res (instantiate-fact-parameters
			(rest actions) predicates)))
	      (dolist (pred predicates)
		(let* ((aname (symnumcat (car (first actions)) '- (car pred)))
		       (ppar (cdr (rename-predicate-parameters pred)))
		       (apar (append
			      (remove-bindings
			       fps (cdr (assoc ':parameters
					       (cdr (first actions)))))
			      ppar))
		       (rpred (cons (car pred) (mapcar #'car ppar)))
		       (aeff (sublis (list (cons (car fps) rpred))
				     (cdr (assoc ':effect
						 (cdr (first actions))))))
		       )
		  (push (cons aname
			      (reassoc ':parameters apar
				       (reassoc ':effect aeff
						(cdr (first actions)))))
			res)))
	      res))
	 )))
   ))

(defun compile-actions (actions intendable-predicates dg)
  (let ((res nil))
    (dolist (act actions)
      (let* ((aname (car act))
	     (aparam (cdr (assoc ':parameters (cdr act))))
	     (actors (cdr (assoc ':actors (cdr act))))
	     (aprec (cdr (assoc ':precondition (cdr act))))
	     (aplist (formula-to-split-lists
		      (cdr (assoc ':precondition (cdr act)))))
	     (aeff (cdr (assoc ':effect (cdr act))))
	     (adds (formula-to-split-lists
		    (cdr (assoc ':effect (cdr act))))))
	(setq res
	 (append res
	  (compile-action aname aparam actors aprec aplist aeff adds nil
			  intendable-predicates dg)))))
    res))

(defun compile-action
  (name param actors prec plist eff adds unj intendable dg)
  (cond
   ((null actors)
    (list (make-action name param prec eff unj)))
   (t (let ((res nil)
	    (counter 1))
	(dolist (add adds)
	  (let ((primary
		 (cond ((eq (car add) 'intends) (caddr add))
		       (t add))))
	    (dolist (intpred intendable)
	      (cond
	       ((and (eq (car primary) intpred)
		     (or (not (eq (car add) 'intends))
			 *allow-delegation-of-intention*))
		(let ((rintpred primary))
		  (setq res
		   (append res
		    (compile-action
		     (symnumcat name '-because- intpred '-A '- counter)
		     param
		     (rest actors)
		     (merge-conjunctions
		      (merge-conjunctions
		       (list 'intends (first actors) rintpred)
		       (non-delegation-conditions
			adds (first actors) primary
			(not (equal primary add)) rintpred intendable))
		      prec)
		     plist
		     (merge-conjunctions
		      (make-conjunction-list
		       (append
			(mapcar
			 #'(lambda (x)
			     (cond ((member (car x) intendable)
				    (list 'justified x
					  (list 'intends
						(first actors) rintpred)))
				   (t nil)))
			 plist)
			(cond ((eq (car add) 'intends)
			       (list
				(list 'delegated (second add) primary)
				(list 'not (list 'delegated
						 (first actors) primary))))
			      (t (list
				  (list 'not (list 'delegated
						   (first actors) primary)))))
			))
		      eff)
		     adds unj intendable dg))))))
	      (cond
	       ((member (car primary) (dg-reachable dg nil (list intpred)))
		(let ((rintpred (rename-predicate-parameters
				 (assoc intpred *predicates*)
				 (length actors))))
		  (setq res
		   (append res
		    (compile-action
		     (symnumcat name '-because- intpred '-B '- counter)
		     (append param (cdr rintpred))
		     (rest actors)
		     (merge-conjunctions
		      (merge-conjunctions
		       (list 'intends (first actors) (strip-types rintpred))
		       (non-delegation-conditions
			adds (first actors) primary
			(not (equal primary add)) rintpred intendable))
		      prec)
		     plist
		     (merge-conjunctions
		      (make-conjunction-list
		       (append
			(mapcar
			 #'(lambda (x)
			     (cond ((member (car x) intendable)
				    (list 'justified x
					  (list 'intends
						(first actors)
						(strip-types rintpred))))
				   (t nil)))
			 plist)
			(cond ((eq (car add) 'intends)
			       (list
				(list 'delegated (second add) primary)
				(list 'not (list 'delegated
						 (first actors) primary))))
			      (t nil))
			))
		      eff)
		     adds
		     (cons
		      (cons primary (list 'intends (first actors)
					  (strip-types rintpred)))
		      unj)
		     intendable dg))))))
	      ))
	  (setq counter (+ counter 1)))
	res))
   ))

(defun make-conjunction-list (clist)
  (cons 'and (remove nil clist)))

(defun non-delegation-conditions
  (adds actor primary primary-is-del intpred intendable)
  (reduce #'merge-conjunctions
	  (mapcar
	   #'(lambda (x)
	       (cond
		((and (eq (car x) 'intends)
		      (not (member (car (third x)) intendable)))
		 nil)
		((and (not (eq (car x) 'intends))
		      (not (member (car x) intendable)))
		 nil)
		((and primary-is-del
		      (eq (car x) 'intends)
		      (equal (third x) intpred)
		      (equal (third x) primary))
		 (list 'forall '(?c - character)
		       (list 'imply
			     (list 'not (list '= '?c actor))
			     (list 'not (list 'delegated '?c (third x))))))
		((and (not primary-is-del)
		      (equal x intpred)
		      (equal x primary))
		 (list 'forall '(?c - character)
		       (list 'imply
			     (list 'not (list '= '?c actor))
			     (list 'not (list 'delegated '?c x)))))
		((eq (car x) 'intends)
		 (list 'forall '(?c - character)
		       (list 'not (list 'delegated '?c (third x)))))
		(t (list 'forall '(?c - character)
			 (list 'not (list 'delegated '?c x))))
		))
	   adds)))

(defun make-action (name param prec eff unj)
  (list name
	(cons ':parameters param)
	(cons ':precondition (translate-modal prec))
	(cons ':effect
	      (translate-modal
	       (merge-conjunctions
		(cons 'and (mapcar
			    #'(lambda (pair)
				(list 'not
				      (list 'justified (car pair)
					    (cdr pair))))
			    unj))
		eff)))
	))

(defun translate-modal (f)
  (cond ((null f) nil)
	((member (car f) '(and or imply not))
	 (cons (car f) (mapcar #'translate-modal (cdr f))))
	((member (car f) '(forall exists))
	 (list (car f) (cadr f) (translate-modal (caddr f))))
	((eq (car f) 'intends)
	 (setq *modal-predicates*
	       (add-to-set (list 'intends (caaddr f))
			   *modal-predicates*))
	 (cons (symnumcat 'intends '- (caaddr f))
	       (cons (cadr f) (cdaddr f))))
	((eq (car f) 'delegated)
	 (setq *modal-predicates*
	       (add-to-set (list 'delegated (caaddr f))
			   *modal-predicates*))
	 (cons (symnumcat 'delegated '- (caaddr f))
	       (cons (cadr f) (cdaddr f))))
	((eq (car f) 'justified)
	 (setq *modal-predicates*
	       (add-to-set (list 'justified
				 (car (second f))
				 (car (third (third f))))
			   *modal-predicates*))
	 (cons (symnumcat 'justified '- (car (second f))
			  '- (car (third (third f))))
	       (append (cdr (second f))
		       (cons (second (third f)) (cdr (third (third f)))))))
	(t f)))

;;; debug: (defun translate-modal (f) f)

;;;;
;; Compile the example scenario

(setq *modal-predicates* nil)
(load-files input-file-name)

(setq s1 (instantiate-fact-parameters
	  *actions* (mapcar
		     #'(lambda (pname) (assoc pname *predicates*))
		     delegatable-predicates)
	  ))
(setq s2 (compile-actions s1 intendable-predicates dg))

(setq *modal-predicates*
      (mapcar
       #'(lambda (mp)
	   (cond ((eq (car mp) 'intends)
		  (cons (symnumcat 'intends- (cadr mp))
			(cons (cons '?intends-who 'character)
			      (cdr (assoc (cadr mp) *predicates*)))))
		 ((eq (car mp) 'delegated)
		  (cons (symnumcat 'delegated- (cadr mp))
			(cons (cons '?del-to-who 'character)
			      (cdr (assoc (cadr mp) *predicates*)))))
		 ((eq (car mp) 'justified)
		  (cons (symnumcat 'justified- (cadr mp) '- (caddr mp))
			(append
			 (cdr (rename-predicate-parameters
			       (assoc (cadr mp) *predicates*) 1))
			 (cons (cons '?intends-who 'character)
			       (cdr (rename-predicate-parameters
				     (assoc (caddr mp) *predicates*) 2))))))
		 ))
       *modal-predicates*))

;;;;
;; Output

(with-open-file
 (stream output-file-name :direction :output)

 (format stream "~&(define (domain npc-jt)~%")
 (format stream "~&  (:requirements :adl :typing :equality)~%~%")
 (format stream "~&  (:types")
 (dolist (typ *types*)
   (format stream "~&   ~a - ~a" (car typ) (cdr typ)))
 (format stream ")~%~%")

 (format stream "~&  (:predicates")
 (dolist (pred *predicates*)
   (format stream "~&   ~a" (cons (car pred) (unparse-typed-list (cdr pred)))))
 (dolist (pred *modal-predicates*)
   (format stream "~&   ~a" (cons (car pred) (unparse-typed-list (cdr pred)))))
 (format stream ")~%~%")

 (dolist (act s2)
   (format stream "  (:action ~a~%" (car act))
   (format stream "    :parameters ~a~%"
	   (unparse-typed-list (cdr (assoc ':parameters (cdr act)))))
   (format stream "    :precondition ~a~%"
	   (cdr (assoc ':precondition (cdr act))))
   (format stream "    :effect ~a)~%~%"
	   (cdr (assoc ':effect (cdr act))))
   )
 (format stream "  )~%")

 ) ; end with-open-file

(quit)
