
(load "/home/path/pkg/inval/inval.lsp")
(setq *pddl-keywords* (reassoc 'actors ':actors *pddl-keywords*))
;; (setq *allow-delegation-of-intention* nil)

(setq input-domain-file-name "ry-story.npddl")
(setq input-problem-file-name "ry-story-1.pddl")
(setq output-domain-file-name "ry-story-mpc.pddl")
(setq output-problem-file-name "ry-story-1-mpc.pddl")
;; (setq intendable-predicates
;;       '(married-to loves controls has dead at))
;; (setq delegatable-predicates
;;       '(loves has dead at))
;; (setq dg '((married-to loves at)
;; 	   (loves controls)
;; 	   (controls has)
;; 	   (has has dead at)
;; 	   (dead at)))

;; (setq input-domain-file-name "ex-story.npddl")
;; (setq input-problem-file-name "ex-story-1.pddl")
;; (setq output-domain-file-name "ex-story-mpc.pddl")
;; (setq output-problem-file-name "ex-story-1-mpc.pddl")
(setq max-intention-depth 4)
(setq intendable-predicates
      '(married-to loves controls has dead at))
(setq delegatable-predicates
      '(loves has dead at))
(setq dg '((married-to loves at)
	   (loves controls)
	   (controls has)
	   (has has dead at loves)
	   (dead at loves)))


(defun basic-achievable-intentions (actions)
  (let ((res nil))
    (dolist (act actions)
      (let ((adds (collect-adds (cdr (assoc ':effect (cdr act)))))
	    (actors (cdr (assoc ':actors (cdr act)))))
	(cond
	 (actors
	  (dolist (add adds)
	    (cond
	     ((not (eq (car (first add)) 'intends))
	      (setq res (add-to-set
			 (cons (car (first add))
			       (remove-duplicates
				(action-parameter-list-to-type-list
				 actors act)))
			 res)))
	     ))))))
    res))

(defun action-parameter-list-to-type-list (params act)
  (let ((decls (cdr (assoc ':parameters (cdr act)))))
    (mapcar #'(lambda (p)
		(let ((ptype (assoc p decls)))
		  (cond (ptype (cdr ptype))
			(t (error "~a is not a parameter of ~a" p act)))))
	    params)))

(defun dg-reachable (dg set new)
  (cond ((endp new) set)
	(t (dg-reachable
	    dg (remove-duplicates (append set (cdr (assoc (first new) dg))))
	    (append (rest new)
		    (remove-if #'(lambda (x) (find x set)) 
			       (cdr (assoc (first new) dg))))))
	))

(defun rename-predicate-parameters (pred &optional (level nil))
  (cons (car pred)
	(mapcar
	 #'(lambda (a)
	     (cond (level
		    (cons (symnumcat (car a) '- (car pred) '- level)
			  (cdr a)))
		   (t (cons (symnumcat (car a) '- (car pred)) (cdr a)))))
	 (cdr pred))))

(defun strip-types (pred)
  (cons (car pred) (mapcar #'car (cdr pred))))

(defun instantiate-fact-parameters (actions predicates)
  (cond
   ((endp actions) nil)
   (t (let ((fps (rassoc-all 'fact (cdr (assoc ':parameters
					       (cdr (first actions)))))))
	(cond
	 ((null fps)
	  (cons (first actions)
		(instantiate-fact-parameters (rest actions) predicates)))
	 ((> (length fps) 1)
	  (error "actions with >1 fact parameters not supported: ~s"
		 (first actions)))
	 (t (let ((res (instantiate-fact-parameters
			(rest actions) predicates)))
	      (dolist (pred predicates)
		(let* ((aname (symnumcat (car (first actions)) '- (car pred)))
		       (ppar (cdr (rename-predicate-parameters pred)))
		       (apar (append
			      (remove-bindings
			       fps (cdr (assoc ':parameters
					       (cdr (first actions)))))
			      ppar))
		       (rpred (cons (car pred) (mapcar #'car ppar)))
		       (aeff (sublis (list (cons (car fps) rpred))
				     (cdr (assoc ':effect
						 (cdr (first actions))))))
		       )
		  (push (cons aname
			      (reassoc ':parameters apar
				       (reassoc ':effect aeff
						(cdr (first actions)))))
			res)))
	      res))
	 )))
   ))

(defun compile-actions (actions intendable-predicates dg)
  (let ((res nil))
    (dolist (act actions)
      (let* ((aname (car act))
	     (aparam (cdr (assoc ':parameters (cdr act))))
	     (actors (cdr (assoc ':actors (cdr act))))
	     (aprec (cdr (assoc ':precondition (cdr act))))
	     (aplist (formula-to-split-lists
		      (cdr (assoc ':precondition (cdr act)))))
	     (aeff (cdr (assoc ':effect (cdr act))))
	     (adds (formula-to-split-lists
		    (cdr (assoc ':effect (cdr act))))))
	(setq res
	 (append
	  res
	  (compile-action
	   aname aparam actors aprec aeff
	   (remove-if
	    #'(lambda (a)
		(cond ((eq (car a) 'intends)
		       (not (member (car (third a)) intendable-predicates)))
		      (t (not (member (car a) intendable-predicates)))
		      ))
	    adds))
	  (mp-actions
	   aname aparam actors aprec aeff
	   (remove-if
	    #'(lambda (a)
	  	(cond ((eq (car a) 'intends)
	  	       (not (member (car (third a)) intendable-predicates)))
	  	      (t (not (member (car a) intendable-predicates)))))
	    adds)
	   (remove-if
	    #'(lambda (a)
	  	(not (member (car a) intendable-predicates)))
	    aplist))
	  ))
	))
    res))

(defun compile-action (name param actors prec eff adds)
  (cond
   ((null actors)
    (list (make-action name param prec eff adds)))
   (t (let ((res nil)
	    (counter 1))
	(dolist (add adds)
	  (let ((primary
		 (cond ((eq (car add) 'intends) (caddr add))
		       (t add))))
	    (setq res
	     (append res
	      (compile-action
	       (symnumcat name '- counter)
	       (append param
		       (list (cons (symnumcat '?c- (length actors)) 'count)
			     (cons (symnumcat '?c- (length actors) '-less-1)
				   'count)))
	       (rest actors)
	       (merge-conjunctions
		(list 'and
		      (list 'intends (first actors) primary)
		      (list 'intention-count (first actors)
			    (symnumcat '?c- (length actors)))
		      (list 'next (symnumcat '?c- (length actors) '-less-1)
			    (symnumcat '?c- (length actors))))
		prec)
	       (merge-conjunctions
		(list 'and
		      (list 'not (list 'intends (first actors) primary))
		      (list 'not (list 'intention-count (first actors)
				       (symnumcat '?c- (length actors))))
		      (list 'intention-count (first actors)
			    (symnumcat '?c- (length actors) '-less-1)))
		eff)
	       adds))))
	  (setq counter (+ counter 1)))
	res))
   ))

(defun mp-actions (name param actors prec eff alist plist)
  (let ((res nil)
	(ac 1)
	(ec 1)
	(pc 1))
    (dolist (actor actors)
      (setq ec 1)
      (dolist (add alist)
	(setq pc 1)
	(dolist (pre plist)
	  (cond
	   ((not (eq (car pre) 'intends))
	    (push
	     (list
	      (symnumcat 'plan-to- name '- ac '- ec '- pc)
	      (cons ':parameters
		    (append (remove-if
			     #'(lambda (p)
				 (not (or (member (car p)
					   (cond ((eq (car add) 'intends)
						  (cdr (third add)))
						 (t (cdr add))))
					  (member (car p) (cdr pre))
					  (eq (car p) actor))))
			     param)
			    (list (cons (symnumcat '?c- ac) 'count)
				  (cons (symnumcat '?c- ac '-plus-1) 'count))))
	      (cons ':precondition
		    (translate-modal
		     (list 'and
			   (cond ((eq (car add) 'intends)
				  (list 'intends actor (third add)))
				 (t (list 'intends actor add)))
			   (list 'intention-count actor (symnumcat '?c- ac))
			   (list 'next (symnumcat '?c- ac)
				 (symnumcat '?c- ac '-plus-1)))))
	      (cons ':effect
		    (translate-modal
		     (list 'and
			   (list 'intends actor pre)
			   (list 'not (list 'intention-count actor
					    (symnumcat '?c- ac)))
			   (list 'intention-count actor
				 (symnumcat '?c- ac '-plus-1))))))
	     res)))
	  (setq pc (+ pc 1)))
	(setq ec (+ ec 1)))
      (setq ac (+ ac 1)))
    res))

(defun make-action (name param prec eff adds)
  (let ((nparam nil)
	(npre nil)
	(neff nil))
    (dolist (add adds)
      (cond
       ((eq (car add) 'intends)
	(setq nparam
	 (append nparam
	  (list (cons (symnumcat '?d- (+ (length nparam) 2)) 'count)
		(cons (symnumcat '?d- (+ (length nparam) 2) '-plus-1) 'count))))
	(setq npre
	 (append npre
	  (list	(list 'intention-count (second add)
		      (symnumcat '?d- (length nparam)))
		(list 'next (symnumcat '?d- (length nparam))
		      (symnumcat '?d- (length nparam) '-plus-1)))))
	(setq neff
	 (append neff
	  (list (list 'not (list 'intention-count (second add)
				 (symnumcat '?d- (length nparam))))
		(list 'intention-count (second add)
		      (symnumcat '?d- (length nparam) '-plus-1)))))
	)))
    (list name
	  (cons ':parameters (append param nparam))
	  (cons ':precondition
		(translate-modal (merge-conjunctions prec (cons 'and npre))))
	  (cons ':effect
		(translate-modal (merge-conjunctions eff (cons 'and neff))))
	  )))

(defun translate-modal (f)
  (cond ((null f) nil)
	((member (car f) '(and or imply not))
	 (cons (car f) (mapcar #'translate-modal (cdr f))))
	((member (car f) '(forall exists))
	 (list (car f) (cadr f) (translate-modal (caddr f))))
	((eq (car f) 'intends)
	 (setq *modal-predicates*
	       (add-to-set (list 'intends (caaddr f))
			   *modal-predicates*))
	 (cons (symnumcat 'intends '- (caaddr f))
	       (cons (cadr f) (cdaddr f))))
	((eq (car f) 'delegated)
	 (setq *modal-predicates*
	       (add-to-set (list 'delegated (caaddr f))
			   *modal-predicates*))
	 (cons (symnumcat 'delegated '- (caaddr f))
	       (cons (cadr f) (cdaddr f))))
	((eq (car f) 'justified)
	 (setq *modal-predicates*
	       (add-to-set (list 'justified
				 (car (second f))
				 (car (third (third f))))
			   *modal-predicates*))
	 (cons (symnumcat 'justified '- (car (second f))
			  '- (car (third (third f))))
	       (append (cdr (second f))
		       (cons (second (third f)) (cdr (third (third f)))))))
	(t f)))

;;; debug: (defun translate-modal (f) f)

;;;;
;; Compile the example scenario

(setq *modal-predicates* nil)
(load-files input-domain-file-name input-problem-file-name)

(setq s1 (instantiate-fact-parameters
	  *actions* (mapcar
		     #'(lambda (pname) (assoc pname *predicates*))
		     delegatable-predicates)
	  ))
(setq s2 (compile-actions s1 intendable-predicates dg))

(setq *modal-predicates*
      (mapcar
       #'(lambda (mp)
	   (cond ((eq (car mp) 'intends)
		  (cons (symnumcat 'intends- (cadr mp))
			(cons (cons '?intends-who 'character)
			      (cdr (assoc (cadr mp) *predicates*)))))
		 ((eq (car mp) 'delegated)
		  (cons (symnumcat 'delegated- (cadr mp))
			(cons (cons '?del-to-who 'character)
			      (cdr (assoc (cadr mp) *predicates*)))))
		 ((eq (car mp) 'justified)
		  (cons (symnumcat 'justified- (cadr mp) '- (caddr mp))
			(append
			 (cdr (rename-predicate-parameters
			       (assoc (cadr mp) *predicates*) 1))
			 (cons (cons '?intends-who 'character)
			       (cdr (rename-predicate-parameters
				     (assoc (caddr mp) *predicates*) 2))))))
		 ))
       *modal-predicates*))

(setq cnlist nil)
(dotimes (i (+ max-intention-depth 1))
  (push (symnumcat 'n i) cnlist))
(setq static-next-list nil)
(dotimes (i max-intention-depth)
  (push (list 'next (symnumcat 'n i) (symnumcat 'n (+ i 1)))
	static-next-list))
	
(setq *types* (append *types* (list (cons 'count 'object))))
(setq *objects*
      (append *objects*
	      (mapcar #'(lambda (obj) (cons obj 'count)) cnlist)))
(setq *init*
      (append *init*
	      (mapcar #'(lambda (char)
			  (list 'intention-count char 'n0))
		      (objects-of-type 'character *types* *objects*))
	      static-next-list))
(setq *goal*
      (merge-conjunctions
       (cons 'and
	     (mapcar #'(lambda (char)
			 (list 'intention-count char 'n0))
		     (objects-of-type 'character *types* *objects*)))
       *goal*))

;;;;
;; Output

(with-open-file
 (stream output-domain-file-name :direction :output)

 (format stream "~&(define (domain npc-mp)~%")
 (format stream "~&  (:requirements :adl :typing :equality)~%~%")
 (format stream "~&  (:types")
 (dolist (typ *types*)
   (format stream "~&   ~a - ~a" (car typ) (cdr typ)))
 (format stream ")~%~%")

 (format stream "~&  (:predicates")
 (format stream "~&   ~a" '(intention-count ?who - character ?c - count))
 (format stream "~&   ~a" '(next ?c - count ?c-next - count))
 (dolist (pred *predicates*)
   (format stream "~&   ~a" (cons (car pred) (unparse-typed-list (cdr pred)))))
 (dolist (pred *modal-predicates*)
   (format stream "~&   ~a" (cons (car pred) (unparse-typed-list (cdr pred)))))
 (format stream ")~%~%")

 (dolist (act s2)
   (format stream "  (:action ~a~%" (car act))
   (format stream "    :parameters ~a~%"
	   (unparse-typed-list (cdr (assoc ':parameters (cdr act)))))
   (format stream "    :precondition ~a~%"
	   (cdr (assoc ':precondition (cdr act))))
   (format stream "    :effect ~a)~%~%"
	   (cdr (assoc ':effect (cdr act))))
   )
 (format stream "  )~%")

 ) ; end with-open-file


(with-open-file
 (stream output-problem-file-name :direction :output)

 (format stream "~&(define (problem ~a)~%" *problem-name*)
 (format stream "~&  (:domain npc-mp)~%~%")
 (format stream "~&  (:objects")
 (dolist (obj *objects*)
   (format stream "~&   ~a - ~a" (car obj) (cdr obj)))
 (format stream ")~%~%")

 (format stream "~&  (:init")
 (dolist (atom *init*)
   (format stream "~&   ~a" atom))
 (format stream ")~%~%")

 (format stream "~&  (:goal ~a)~%" *goal*)
 (format stream "  )~%")

 ) ; end with-open-file

(quit)
